from typing import Tuple

from ecdsa import SECP256k1
from ecdsa.curves import Curve

from . import jacobian_coordinates
from .coding import (encode_number,
                     hex_string_to_int)
from .hashes import keccak_256_hash
from .utils import inversion


def verifying_key(signature: str,
                  message: str) -> bytes:
    prepended_message = ('\x19Ethereum Signed Message:\n'
                         + str(len(message))
                         + message)
    message_hash = (keccak_256_hash(prepended_message.encode('ascii'))
                    .hexdigest())
    v, r, s = decode_signature(signature)

    return verifying_key_from_hash(message_hash,
                                   v=v,
                                   r=r,
                                   s=s)


def verifying_key_from_hash(message_hash: str,
                            *,
                            v: int,
                            r: int,
                            s: int) -> bytes:
    left, right = verifying_key_pair(message_hash,
                                     v=v,
                                     r=r,
                                     s=s)
    result = (encode_number(left, 256, 32)
              + encode_number(right, 256, 32))

    if len(result) != 64:
        err_msg = ('Invalid public key, '
                   'should have length of 64')
        raise ValueError(err_msg)

    return result


def verifying_key_pair(message_hash: str,
                       *,
                       v: int,
                       r: int,
                       s: int,
                       curve: Curve = SECP256k1) -> Tuple[int, int]:
    if not (27 <= v <= 34):
        raise ValueError(f'"v" must in range 27-34, '
                         f'but found {v}')

    curve_fp = curve.curve
    p = curve_fp.p()
    a = curve_fp.a()
    b = curve_fp.b()
    n = curve.order
    g_x = curve.generator.x()
    g_y = curve.generator.y()

    x = r
    beta = pow((x ** 3 + a * x + b) % p, (p + 1) // 4, p)
    y = beta if v % 2 ^ beta % 2 else p - beta
    z = hex_string_to_int(message_hash)
    g_z = jacobian_coordinates.multiply((g_x, g_y, 1), (n - z) % n)
    xy = jacobian_coordinates.multiply((x, y, 1), s)
    q_r = jacobian_coordinates.add(g_z, xy)
    q = jacobian_coordinates.multiply(q_r, inversion(r, n))
    return jacobian_coordinates.to_affine(q)


def decode_signature(signature: str) -> Tuple[int, int, int]:
    r = hex_string_to_int(signature[:64])
    s = hex_string_to_int(signature[64:128])
    v = hex_string_to_int(signature[128:])
    return v, r, s



from functools import partial
from hashlib import sha3_256
from typing import (Union,
                    Type)

from ecdsa import (SECP256k1,
                   SigningKey,
                   VerifyingKey)

from .hashes import keccak_256_hash


def hex_bytes_to_key(key_cls: Union[Type[SigningKey],
                                    Type[VerifyingKey]],
                     hex_bytes: bytes,
                     *,
                     curve=SECP256k1,
                     hash_function=sha3_256) -> SigningKey:
    return key_cls.from_string(hex_bytes,
                               curve=curve,
                               hashfunc=hash_function)


hex_bytes_to_signing_key = partial(hex_bytes_to_key,
                                   SigningKey)
hex_bytes_to_verifying_key = partial(hex_bytes_to_key,
                                     VerifyingKey)


def signing_key_to_address(key: SigningKey) -> str:
    # based on
    # https://www.reddit.com/r/ethereum/comments/69qfkv/python_script_that_converts_ethereum_private_key/
    verifying_key = key.get_verifying_key()
    return verifying_key_to_address(verifying_key)


def verifying_key_to_address(key: VerifyingKey) -> str:
    key_string = key.to_string()
    key_hash = keccak_256_hash(key_string)
    key_hash_hex = key_hash.hexdigest()
    return key_hash_hex[24:]



import codecs
import string
from functools import partial

code_strings = {
    2: '01',
    8: string.octdigits,
    10: string.digits,
    16: string.digits + 'abcdef',
    32: 'abcdefghijklmnopqrstuvwxyz234567',
    58: '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz',
    256: ''.join(map(chr, range(256)))
}


def encode_number(number: int,
                  base: int,
                  min_length: int) -> bytes:
    try:
        code_string = code_strings[base]
    except KeyError:
        raise ValueError(f'Invalid base: {base}.')

    result = bytes()
    while number > 0:
        code = code_string[number % base]
        result = bytes([ord(code)]) + result
        number //= base

    pad_size = min_length - len(result)

    padding_element = b'\x00' if base == 256 else b'1' if base == 58 else b'0'
    if pad_size > 0:
        result = padding_element * pad_size + result

    return result


def hex_string_to_int(hex_string: str) -> int:
    return int(hex_string, 16)


def big_endian_to_int(big_endian: bytes) -> int:
    return int.from_bytes(big_endian.lstrip(bytes([0])),
                          byteorder='big')


decode_hex = partial(codecs.decode,
                     encoding='hex')



from typing import (Union,
                    Dict,
                    Tuple)

from Cryptodome.Cipher import AES
from Cryptodome.Util import Counter

from . import aes
from .coding import (big_endian_to_int,
                     decode_hex)
from .hashes import (keccak_256_hash,
                     scrypt_hash)


def decode_keystore_json(key_json: Dict[str, Union[str, int, dict]],
                         password: str) -> bytes:
    crypto = key_json['crypto']
    aes_key, mac_bytes_head = derived_key_parts(password=password,
                                                kdf_params=crypto['kdfparams'])
    cipher_text = decode_hex(crypto['ciphertext'])
    validate_mac(cipher_text=cipher_text,
                 mac_bytes_head=mac_bytes_head,
                 predicted_mac=decode_hex(crypto['mac']))

    cipher_params = crypto['cipherparams']
    initial_value = initial_value_str_to_int(string=cipher_params['iv'])
    counter = Counter.new(nbits=128,
                          initial_value=initial_value,
                          allow_wraparound=True)
    return aes.decrypt(text=cipher_text,
                       key=aes_key,
                       counter=counter,
                       mode=AES.MODE_CTR)


def derived_key_parts(*,
                      password: str,
                      kdf_params: Dict[str, Union[int, str]]
                      ) -> Tuple[bytes, bytes]:
    derived_key = scrypt_hash(password,
                              salt=decode_hex(kdf_params['salt']),
                              buffer_length=kdf_params['dklen'],
                              p=kdf_params['p'],
                              r=kdf_params['r'],
                              n=kdf_params['n'])
    if len(derived_key) < 32:
        raise ValueError('Derived key must be at least 32 bytes long')
    return derived_key[:16], derived_key[16:32]


def validate_mac(*,
                 cipher_text: bytes,
                 mac_bytes_head: bytes,
                 predicted_mac: bytes) -> None:
    mac_bytes = mac_bytes_head + cipher_text
    mac = keccak_256_hash(mac_bytes).digest()
    if mac != predicted_mac:
        raise ValueError('MAC mismatch. '
                         'Password seems to be incorrect.')


def initial_value_str_to_int(string: str) -> int:
    return big_endian_to_int(decode_hex(string))



from typing import Tuple

from ecdsa import SECP256k1

from .utils import inversion

VectorType = Tuple[int, int, int]


def multiply(vector: VectorType,
             multiplier: int,
             *,
             order=SECP256k1.order) -> VectorType:
    if vector[1] == 0 or multiplier == 0:
        return 0, 0, 1
    if multiplier == 1:
        return vector
    if multiplier < 0 or multiplier >= order:
        return multiply(vector, multiplier % order)

    quotient, remainder = divmod(multiplier, 2)
    if remainder == 0:
        return double(multiply(vector, quotient))
    if remainder == 1:
        return add(double(multiply(vector,
                                   quotient)),
                   vector)


def add(vector: VectorType,
        other_vector: VectorType,
        *,
        p=SECP256k1.curve.p()) -> VectorType:
    """
    Bernstein-Lange addition

    more info at
    https://www.hyperelliptic.org/EFD/g1p/auto-shortw-jacobian.html#addition-add-2007-bl
    """
    vector_x, vector_y, vector_z = vector
    other_vector_x, other_vector_y, other_vector_z = other_vector

    if not vector_y:
        return other_vector
    if not other_vector_y:
        return vector

    u_1 = (vector_x * other_vector_z ** 2) % p
    u_2 = (other_vector_x * vector_z ** 2) % p
    s_1 = (vector_y * other_vector_z ** 3) % p
    s_2 = (other_vector_y * vector_z ** 3) % p

    if u_1 == u_2:
        if s_1 != s_2:
            return 0, 0, 1
        return double(vector)

    h = u_2 - u_1
    r = s_2 - s_1
    h_2 = (h ** 2) % p
    j = (h * h_2) % p
    v = (u_1 * h_2) % p
    result_x = (r ** 2 - j - 2 * v) % p
    result_y = (r * (v - result_x) - s_1 * j) % p
    result_z = (h * vector_z * other_vector_z) % p
    return result_x, result_y, result_z


def double(vector,
           *,
           a: int = SECP256k1.curve.a(),
           p: int = SECP256k1.curve.p()) -> VectorType:
    """
    Chudnovsky–Chudnovsky doubling

    more info at
    http://www.hyperelliptic.org/EFD/g1p/auto-shortw-jacobian.html#doubling-dbl-1986-cc
    """
    vector_x, vector_y, vector_z = vector
    if not vector_y:
        return 0, 0, 0
    squared_vector_y = (vector_y ** 2) % p
    s = (4 * vector_x * squared_vector_y) % p
    m = (3 * vector_x ** 2 + a * vector_z ** 4) % p
    result_x = (m ** 2 - 2 * s) % p
    result_y = (m * (s - result_x) - 8 * squared_vector_y ** 2) % p
    result_z = (2 * vector_y * vector_z) % p
    return result_x, result_y, result_z


def to_affine(vector: VectorType,
              *,
              p: int = SECP256k1.curve.p()) -> Tuple[int, int]:
    vector_x, vector_y, vector_z = vector
    z = inversion(vector_z, p)
    return (vector_x * z ** 2) % p, (vector_y * z ** 3) % p



from typing import (Union,
                    Dict)

from Cryptodome.Cipher import AES


def decrypt(*,
            text: str,
            key: bytes,
            counter: Dict[str, Union[int, bool, bytes]] = None,
            mode: int = AES.MODE_CTR) -> bytes:
    encryptor = AES.new(key,
                        mode,
                        counter=counter)
    return encryptor.decrypt(text)



from functools import partial

import scrypt
from Cryptodome.Hash import keccak


def keccak_hash(target_bytes: bytes,
                *,
                digest_bits: int) -> keccak.Keccak_Hash:
    return keccak.new(digest_bits=digest_bits,
                      data=target_bytes)


keccak_256_hash = partial(keccak_hash,
                          digest_bits=256)


def scrypt_hash(password: str,
                *,
                salt: bytes,
                n: int,
                r: int,
                p: int,
                buffer_length: int) -> bytes:
    return scrypt.hash(password,
                       salt,
                       n, r, p,
                       buffer_length)



def inversion(a: int, n: int) -> int:
    if a == 0:
        return 0
    lm, hm = 1, 0
    low, high = a % n, n
    while low > 1:
        r = high // low
        nm, new = hm - lm * r, high - low * r
        lm, low, hm, high = nm, new, lm, low
    return lm % n



from setuptools import (setup,
                        find_packages)

setup(name='ethereum_helpers',
      packages=find_packages(),
      version='0.0.0',
      description='Helper functions for working with ethereum network.',
      author='Azat Ibrakov',
      author_email='azatibrakov@gmail.com',
      url='https://github.com/lycantropos/ethereum_helpers',
      download_url='https://github.com/lycantropos/ethereum_helpers/archive/'
                   'master.tar.gz',
      keywords=['ethereum'],
      install_requires=[
          'ecdsa>=0.13',
          'scrypt>=0.8.0',
          'pycryptodomex>=3.4.6',
      ])
